const Knex = require("knex");
const dayjs = require("dayjs");
const config = require("config");

/**
 * @typedef {Object} ReceiptsDB
 * @property {number} id
 * @property {string} receipt
 * @property {number} payment_id
 * @property {string} created_at
 */
/** @returns {Knex.QueryBuilder<ReceiptsDB, {}>} */
const Receipts = () => __database_client("receipts");

class Receipt {
  /**
   *
   * @param {Object} data
   * @param {number} data.id
   * @param {string} data.company
   * @param {string} data.name
   * @param {string} data.email
   * @param {string} data.address
   * @param {string} data.receipt
   * @param {number} data.payment_id
   * @param {string} data.created_at
   */
  constructor(data) {
    this.id = data.id;
    this.public_id = "R" + (data.id + config.get("price.number_range.invoices"));
    if (data.id < 0) {
      this.public_id = "TEST_INVOICE";
    }
    this.company = data.company;
    this.name = data.name;
    this.email = data.email;
    this.address = data.address;
    this.receipt = data.receipt;
    this.payment_id = data.payment_id;
    this.created_at = dayjs(data.created_at);
  }

  /**
   * @param {Object} data
   * @param {number} data.id
   * @param {string} data.company
   * @param {string} data.name
   * @param {string} data.email
   * @param {string} data.address
   * @param {string} [data.receipt]
   * @param {number} data.payment_id
   * @param {string} data.created_at
   *
   * @returns {Promise<Receipt>}
   */
  static async CREATE_NEW_RECEIPT(data) {
    data.created_at = dayjs().format("YYYY-MM-DD HH:mm:ss");
    return Receipts()
      .insert(data, "*")
      .then((result) => {
        if (result.length !== 1) {
          return Promise.reject("Error creating Receipt");
        }
        return new Receipt(result[0]);
      });
  }

  /**
   * 
   * @param {number} internal_id 
   * @returns {Promise<Receipt>}
   */
  static async LOAD_RECEIPT_FROM_INTERNAL_ID(internal_id) {
    return Receipts().where("id", internal_id).first().then(receipt_data => {
      if (receipt_data !== undefined) {
        return new Receipt(receipt_data);
      }
      return Promise.reject("Cannot find Receipt with id " + internal_id);
    })
  }

  /**
   *
   * @param {string} receiptdata
   *
   * @returns {Promise<Receipt>}
   */
  async attachReceipt(receiptdata) {
    return Receipts()
      .where("id", this.id)
      .update({ receipt: receiptdata }, "*")
      .then((result) => {
        if (result && result.length === 1) {
          return new Receipt(result[0]);
        }
        return Promise.reject("Failed to insert receipt");
      });
  }
}

module.exports = Receipt;
